// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Text;
using System.Xml;
using System.Collections.Generic;
using PiDuce.Common; // remove me!
using System.Diagnostics;

namespace PiDuce.Common
{
	public interface ILabelSet
	{
		ISet<string> Labels { get; }
		bool Has(string label);
		bool IsEmpty();
		bool IsSubsetOf(ILabelSet labels);
		bool Includes(ILabelSet labels);
		bool Intersects(ILabelSet labels);
		void ToXml(XmlWriter writer);
	}

	public abstract class LabelSet : ILabelSet
	{
    protected ISet<string> labels;

    public ISet<string> Labels { get { return labels; } }

		protected LabelSet()
		{ this.labels = new ArraySet<string>(); }

    protected LabelSet(ISet<string> labels)
    { this.labels = new ArraySet<string>(labels); }

		protected LabelSet(string label)
		{
      this.labels = new ArraySet<string>();
			this.labels.Add(label);
		}

		public abstract bool Has(string labeld);
		public abstract bool IsEmpty();
		public abstract void ToXml(XmlWriter writer);

    private static ISet<string> Union(ISet<string> s1, ISet<string> s2)
		{
      ISet<string> res = new ArraySet<string>(s1);
			foreach (string s in s2)
				if (!res.Contains(s))
					res.Add(s);
			return res;
		}

    private static ISet<string> Difference(ISet<string> s1, ISet<string> s2)
		{
      ISet<string> res = new ArraySet<string>();
			foreach (string s in s1)
				if (!s2.Contains(s))
					res.Add(s);
			return res;
		}

    private static ISet<string> Intersection(ISet<string> s1, ISet<string> s2)
		{
      ISet<string> res = new ArraySet<string>();
			foreach (string s in s1)
				if (s2.Contains(s))
					res.Add(s);
			return res;
		}

		public static ILabelSet Union(ILabelSet s1, ILabelSet s2)
		{
			if (s1 is WithLabels && s2 is WithLabels)
				return new WithLabels(Union(s1.Labels, s2.Labels));
			else if (s1 is WithLabels && s2 is WithoutLabels)
				return new WithoutLabels(Difference(s2.Labels, s1.Labels));
			else if (s1 is WithLabels && s2 is WithLabels)
				return new WithoutLabels(Difference(s1.Labels, s2.Labels));
			else
				return new WithoutLabels(Intersection(s1.Labels, s2.Labels));
		}

		public static ILabelSet Difference(ILabelSet s1, ILabelSet s2)
		{
			if (s1 is WithLabels && s2 is WithLabels)
				return new WithLabels(Difference(s1.Labels, s2.Labels));
			else if (s1 is WithoutLabels && s2 is WithLabels)
        return new WithoutLabels(Union(s1.Labels, s2.Labels));
			else if (s1 is WithLabels && s2 is WithoutLabels)
				return new WithLabels(Intersection(s1.Labels, s2.Labels));
			else
				return new WithLabels(Difference(s2.Labels, s1.Labels));
		}

		public static ILabelSet Intersection(ILabelSet s1, ILabelSet s2)
		{
			if (s1 is WithLabels && s2 is WithLabels)
				return new WithLabels(Intersection(s1.Labels, s2.Labels));
			else if (s1 is WithoutLabels && s2 is WithLabels)
				return new WithLabels(Difference(s2.Labels, s1.Labels));
			else if (s1 is WithLabels && s2 is WithoutLabels)
				return new WithLabels(Difference(s1.Labels, s2.Labels));
			else
				return new WithoutLabels(Union(s1.Labels, s2.Labels));
		}

		public static ILabelSet Complement(ILabelSet s)
		{
			if (s is WithLabels)
				return new WithoutLabels(s.Labels);
			else
				return new WithLabels(s.Labels);
		}

		public static ILabelSet Singleton(string s)
		{ return new WithLabels(s); }

		public static ILabelSet Empty()
		{ return new WithLabels(); }

		public static ILabelSet Any()
		{ return new WithoutLabels(); }

    public static ILabelSet With(ISet<string> labels)
		{ return new WithLabels(labels); }

    public static ILabelSet Without(ISet<string> labels)
		{ return new WithoutLabels(labels); }

		public bool IsSubsetOf(ILabelSet s)
		{ return Difference(this, s).IsEmpty(); }

		public bool Includes(ILabelSet s)
		{ return s.IsSubsetOf(this); }

		public bool Intersects(ILabelSet s)
		{ return !Intersection(this, s).IsEmpty(); }

		protected void ToStringAux(StringBuilder builder)
		{
      foreach (string label in Labels)
        builder.Append('`' + label + ',');
      builder.Replace(",", "", builder.Length - 1, 1);
		}

		protected void ToXmlAux(XmlWriter writer)
		{
			foreach (string l in Labels) {
				writer.WriteStartElement("label");
				writer.WriteAttributeString("name", l);
				writer.WriteEndElement(); // </label>
			}
		}

		public static ILabelSet FromXml(XmlNode node)
		{
      ISet<string> labels = new ArraySet<string>();
			foreach (XmlNode label in node.ChildNodes)
				labels.Add(label.Attributes["name", ""].Value);
			switch (node.LocalName) {
				case "with":
					return With(labels);
				case "without":
					return Without(labels);
				default:
					throw new ApplicationException("Invalid labels");
			}
		}
	}

	public class WithLabels : LabelSet
	{
		public WithLabels() : base()
		{ }

		public WithLabels(string label) : base(label)
		{ }

    public WithLabels(ISet<string> labels) : base(labels)
		{ }

		public override bool Has(string label)
		{ return Labels.Contains(label); }

		public override bool IsEmpty()
		{ return Labels.Count == 0; }

		public override string ToString()
		{
      StringBuilder builder = new StringBuilder();
      if (Labels.Count != 1) builder.Append("{");
      ToStringAux(builder);
      if (Labels.Count != 1) builder.Append("}");
      return builder.ToString();
		}

		public override void ToXml(XmlWriter writer)
		{
			writer.WriteStartElement("with");
			ToXmlAux(writer);
			writer.WriteEndElement(); // </with>
		}
	}

	public class WithoutLabels : LabelSet
	{
		public WithoutLabels() : base()
		{ }

		public WithoutLabels(string label) : base(label)
		{ }

    public WithoutLabels(ISet<string> labels) : base(labels)
		{ }

		public override bool Has(string label)
		{ return !Labels.Contains(label); }

		public override bool IsEmpty()
		{ return false; }

		public override string ToString()
		{
			if (Labels.Count == 0)
				return "*";
			else {
				StringBuilder builder = new StringBuilder();
				builder.Append("{");
				builder.Append("*");
				builder.Append("\\");
				ToStringAux(builder);
				builder.Append("}");
				return builder.ToString();
			}
		}

		public override void ToXml(XmlWriter writer)
		{
			writer.WriteStartElement("without");
			ToXmlAux(writer);
			writer.WriteEndElement(); // </without>
		}
	}
}
