// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections.Generic;
using System.Collections;

namespace PiDuce.Common
{
	public interface ISet<T> : ICollection<T>
	{
		bool IsEmpty();
		bool IsSubsetOf(ISet<T> s);
	}

	public class ArraySet<T> : ISet<T>
	{
		private readonly IList<T> content;

		public ArraySet()
		{ content = new List<T>(); }

    public int Count { get { return content.Count; } }

    public bool IsReadOnly { get { return content.IsReadOnly; } }

		public ArraySet(T element)
		{
			content = new List<T>();
			content.Add(element);
		}

		public ArraySet(IEnumerable<T> elements)
		{ content = new List<T>(elements); }

		public void Add(T element)
		{
			if (!content.Contains(element))
				content.Add(element);
		}

		public bool Remove(T element)
		{	return content.Remove(element); }

		public bool Contains(T element)
		{ return content.Contains(element); }

		public bool IsEmpty()
		{ return content.Count == 0; }

		public bool IsSubsetOf(ISet<T> s)
		{
			foreach (T element in content)
				if (!s.Contains(element))
					return false;
			return true;
		}

    public IEnumerator<T> GetEnumerator()
    { return content.GetEnumerator(); }

    IEnumerator System.Collections.IEnumerable.GetEnumerator()
    { return (IEnumerator) content.GetEnumerator(); }

    public void Clear()
    { content.Clear(); }

    public void CopyTo(T[] array, int index)
    { content.CopyTo(array, index); }

		public static ArraySet<T> Empty()
		{ return new ArraySet<T>(); }

		public static ArraySet<T> Singleton(T element)
		{ return new ArraySet<T>(element); }

		public static ArraySet<T> Union(ArraySet<T> s1, ArraySet<T> s2)
		{
			ArraySet<T> res = new ArraySet<T>(s1.content);
			foreach (T element in s2.content)
				res.Add(element);
			return res;
		}

		public static ArraySet<T> Intersection(ArraySet<T> s1, ArraySet<T> s2)
		{
			ArraySet<T> res = Empty();
			foreach (T element in s1.content)
				if (s2.Contains(element))
					res.Add(element);
			return res;
		}

		public static ArraySet<T> Difference(ArraySet<T> s1, ArraySet<T> s2)
		{
			ArraySet<T> res = Empty();
			foreach (T element in s1.content)
				if (!s2.Contains(element))
					res.Add(element);
			return res;
		}

		public override bool Equals(Object obj)
		{
			if (obj is ISet<T>) {
				ISet<T> s = (ISet<T>) obj;
				return this.IsSubsetOf(s) && s.IsSubsetOf(this);
			} else
				return false;
		}

		public override int GetHashCode()
		{
			int hash = 0;
			foreach (T element in content)
				hash ^= element.GetHashCode();
			return hash;
		}
	}

#if false
	public class Set<T> : ISet<T>
	{
		private Node node;
		private readonly IComparer<T> comparer;

		private class Node
		{
			public readonly T content;
			public readonly Node left;
			public readonly Node right;
			public readonly int height;

			private Node(T content, Node left, Node right, int height)
			{
				this.content = content;
				this.left = left;
				this.right = right;
				this.heigth = height;
			}

			private static int Height(Node node)
			{ return (node == null) ? 0 : node.height; }

			public static Node Create(T content, Node left, Node right)
			{ return new Node(content, left, right, max(Height(left), Height(right)) + 1); }

			public static Node Create(T content, Node left, Node right, int height)
			{ return new Node(content, left, right, height); }

			public static Node Bal(T content, Node left, Node right)
			{
				int hl = Height(left);
				int hr = Height(right);
				if (hl > hr + 2) {
					Debug.Assert(left != null);
					if (Height(left.left) >= Height(left.right))
						return Create(content, left.content, Create(content, left.right, right));
					else {
						Debug.Assert(left.right != null);
						return Create(left.right.content, Create(left.content, left.left, left.right.left), Create(content, left.right.right, right));
					}
				} else if (hr > hl + 2) {
					Debug.Assert(right != null);
					if (Height(right.right) >= Height(right.left))
						return Create(right.content, Create(content, left, right.left), right.right);
					else {
						Debug.Assert(right.left != null);
						return Create(right.left.content, Create(content, left, right.left.left), Create(right.content, right.left.right, right.right));
					}
				} else
					return Create(content, left, right);
			}
		}

		private static Node AddAux(T x, Node node)
		{
			if (node == null)
				node = Node.Create(x, null, null);
			else {
				switch (x.Compare(node.content)) {
					case 0:
						break;
					case -1:
						node = Node.Bal(node.content, Add(x, node.left), node.right);
						break;
					case 1:
						node = Node.Bal(node.content, node.left, Add(x, node.right));
						break;
					default:
						Debug.Assert(false);
				}
			}
		}

		private static Node Join(T content, Node left, Node right)
		{
			if (left == null)
				return AddAux(content, right);
			else if (right == null)
				return AddAux(content, left);
			else if (left.height > right.height + 2)
				return Node.Bal(left.content, left.left, Join(content, left.right, right));
			else if (right.height > left.height + 2)
				return Node.Bal(right.content, Join(content, left, right.left), right.right);
			else
				return Node.Create(content, left, right);
		}

		private static T MinElt(Node node)
		{
			if (node == null)
				throw new NotFound();
			else if (node.left == null)
				return node.content;
			else
				return MinElt(node.left);
		}

		private static T MaxElt(Node node)
		{
			if (node == null)
				throw new NotFound();
			else if (node.right == null)
				return node.content;
			else
				return MaxElt(node.right);
		}

		private static Node RemoveMinElt(Node node)
		{
			Debug.Assert(node != null);
			if (node.left == null)
				return node.right;
			else
				return Node.Bal(node.content, RemoveMinElt(node.left), node.right);
		}

		private static Node Merge(Node node1, Node node2)
		{
			if (node1 == null)
				return node2;
			else if (node2 == null)
				return node1;
			else
				return Node.Bal(node1, MinElt(node2), RemoveMinElt(node2));
		}

		private static Node Concat(Node node1, Node node2)
		{
			if (node1 == null)
				return node2;
			else if (node2 == null)
				return node1;
			else
				return Join(node1, MinElt(node2), RemoveMinElt(node2));
		}

		private static bool Split(T x, Node node, out Node left, out Node right)
		{
			if (node == null) {
				left = null;
				right = null;
				return false;
			} else {
				int c = x.CompareTo(node.content);
				if (c == 0) {
					left = node.left;
					right = node.right;
					return true;
				} else if (c < 0) {
					Node rl;
					bool present = Split(x, node.left, out left, out rl);
					right = Join(node.content, rl, node.right);
					return present;
				} else {
					Node lr;
					bool present = Split(x, node.right, out lr, out right);
					left = Join(node.content, node.left, lr);
					return present;
				}
			}
		}

		private static bool MemAux(T x, Node node)
		{
			if (node == null)
				return false;
			else {
				int c = x.CompareTo(node.content);
				if (c == 0)
					return true;
				else if (c < 0)
					return MemAux(x, node.left);
				else
					return MemAux(x, node.right);
			}
		}

		private static Node RemoveAux(T x, Node node)
		{
			if (node == null)
				return null;
			else {
				int c = x.CompareTo(node.content);
				if (c == 0)
					return Merge(node.left, node.right);
				else if (c < 0)
					return Bal(node.content, RemoveAux(x, node.left), node.right);
				else
					return Bal(node.content, node.left, RemoveAux(x, node.right));
			}
		}

		private static Node UnionAux(Node s1, Node s2)
		{
			if (s1 == null)
				return s2;
			else if (s2 == null)
				return s1;
			else {
				if (s1.height >= s2.height) {
					if (s2.height == 1)
						return Add(s2.content, s1);
					else {
						Node l2;
						Node r2;
						Split(s1.content, s2, out l2, out r2);
						return Join(s1.content, UnionAux(s1.left, l2), UnionAux(s1.right, r2));
					}
				} else if (s1.height == 1)
					return Add(s1.content, s2);
				else {
					Node l1;
					Node r1;
					Split(s2.content, s1, out l1, out r1);
					return Join(s2.content, UnionAux(l1, s2.left), UnionAux(r1, s2.right));
				}
			}
		}

		private static Node InterAux(Node s1, Node s2)
		{
			if (s1 == null || s2 == null)
				return null;
			else {
				Node l2;
				Node r2;
				if (Split(s1.content, s2, out l2, out r2))
					return Concat(InterAux(s1.left, l2), InterAux(s1.right, r2));
				else
					return Join(s1.content, InterAux(s1.left, l2), InterAux(s1.right, r2));
			}
		}

		private static Node DiffAux(Node s1, Node s2)
		{
			if (s1 == null)
				return null;
			else if (s2 == null)
				return s1;
			else {
				Node l2;
				Node r2;
				if (Split(s1.content, s2, out l2, out r2))
					return Join(s1.content, DiffAux(s1.left, l2), DiffAux(s1.right, r2));
				else
					return Concat(DiffAux(s1.left, l2), DiffAux(s1.right, r2));
			}
		}

		private class More
		{
			private T elt;
			private Node t;
			private More next;

			public More(T elt, Node t, More next)
			{
				this.elt = elt;
				this.t = t;
				this.next = next;
			}
		}

		private static More ConsEnum(Node s, More e)
		{
			if (s == null)
				return e;
			else
				return ConsEnum(node.left, new More(node.content, node.right, e));
		}

		private static int CompareAux(More e1, More e2)
		{
			if (e1 == null && e2 == null)
				return 0;
			else if (e1 == null)
				return -1;
			else if (e2 == null)
				return 1;
			else {
				int c = e1.elt.CompareTo(e2.elt);
				if (c != 0)
					return c;
				else
					return CompareAux(ConsEnum(e1.t, e1.next), ConsEnum(e2.t, e2.next));
			}
		}

		private static bool SubsetAux(Node s1, Node s2)
		{
			if (s1 == null)
				return true;
			else if (s2 == null)
				return false;
			else {
				int c = s1.content.CompareTo(s2.content);
				if (c == 0)
					return SubsetAux(s1.left, s2.left) && SubsetAux(s1.right, s2.right);
				else if (c < 0)
					return SubsetAux(new Node(s1.content, s1.left, null, 0), s2.left)
						&& SubsetAux(s1.right, s2);
				else
					return SubsetAux(new Node(s1.content, null, s1.right, 0), s2.right)
						&& SubsetAux(s1.left, s2);
			}
		}

		public static Set<T> Empty()
		{ return new Set(); }

		public bool IsEmpty()
		{ return node == null; }

		public bool Contains(T x)
		{ return MemAux(node, x); }

		public static Set<T> Singleton(T x)
		{
			Set s = Empty();
			s.Add(x);
			return s;
		}

		public void Add(T x)
		{ node = AddAux(x, node); }

		public void Remove(T x)
		{ node = RemoveAux(x, node); }

		public void Union(Set<T> s)
		{ node = UnionAux(node, s.node); }

		public void Intersection(Set<T> s)
		{ node = InterAux(node, s.node); }

		public void Difference(Set<T> s)
		{ node = DiffAux(node, s.node); }

		public int Compare(Set<T> s)
		{ return CompareAux(ConsEnum(node, null), ConsEnum(s.node, null)); }

		public bool Equals(Set<T> s)
		{ return Compare(s) == 0; }

		public bool SubsetOf(Set<T> s)
		{ return SubsetAux(node, s.node); }
	}
#endif
}

