// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Xml;
using PiDuce.Common;

using IPattern = PiDuce.Common.ISchema;

namespace PiDuce.Machine
{
  public class JoinInputRequest : IInputRequest
  {
    private JoinInputDefinition definition;
    private Channel channel;
    private int index;

    public JoinInputRequest(JoinInputDefinition definition, Channel channel, int index)
    {
      this.definition = definition;
      this.channel = channel;
      this.index = index;
    }

    public bool Send(IValue v)
    { 
			return definition.Send(this, v);
		}

    public void Empty()
    { definition.Empty(this); }

		public Channel Channel { get { return channel; } }
		public int Index { get { return index; } }
  }

	public interface IJoinAutomaton
	{
		void Full(int channelIndex);
		void Empty(int channelIndex);
		int ActiveRow { get; }
	}

	public class NaiveJoinAutomaton : IJoinAutomaton
	{
		private IList<JoinInputDefinition.Row> rows;
    private IDictionary<int, bool> status;

    public NaiveJoinAutomaton(IList<JoinInputDefinition.Row> rows)
    {
			this.rows = rows;
      this.status = new Dictionary<int, bool>();
    }

		public int ActiveRow {
			get {
				for (int i = 0; i < rows.Count; i++) {
					bool full = true;
					foreach (JoinInputDefinition.Atom atom in rows[i].Atoms) {
				  	if (!status.ContainsKey(atom.Index)) {
				    	full = false;
				    	break;
				  	}
					}
		
					if (full)
				  	return i;
      	}

				return -1;
			}
		}

    public void Full(int channelIndex)
    {
      if (!status.ContainsKey(channelIndex))
				status.Add(channelIndex, true);
    }

    public void Empty(int channelIndex)
    {
      if (status.ContainsKey(channelIndex))
				status.Remove(channelIndex);
    }
	}

	public class JoinAutomaton : IJoinAutomaton
	{
    private int state;
    private int nChannels;
    private int[,] transitions;
    private int[] continuations;

		private class State
		{
			public bool Initialized;
			public int Index;
			public int Final;
			public int[] Transitions;
			private bool[] channels;

			public State(int nChannels)
			{
				this.Initialized = false;
				this.Index = -1;
				this.Final = -1;
				this.Transitions = null;
				Debug.Assert(nChannels > 0);
				this.channels = new bool[nChannels];
				for (int i = 0; i < nChannels; i++)
					this.channels[i] = false;
			}

			private State(bool[] channels)
			{
				this.channels = (bool[]) channels.Clone();
			}

			private State(bool[] channels, int toggle)
			{
				Debug.Assert(toggle >= 0 && toggle < channels.Length);
				this.channels = (bool[]) channels.Clone();
				this.channels[toggle] = !this.channels[toggle];
			}

			public override bool Equals(object obj)
			{
				Debug.Assert(obj is State);

				State state = (State) obj;
				if (channels.Length != state.channels.Length)
					return false;

				for (int i = 0; i < channels.Length; i++)
					if (channels[i] != state.channels[i])
						return false;

				return true;
			}

			public override int GetHashCode()
			{ return base.GetHashCode(); }

			public State Full(int s)
			{
				if (channels[s])
					return this;
				else
					return new State(channels, s);
			}

			public State Empty(int s)
			{
				if (!channels[s])
					return this;
				else
					return new State(channels, s);
			}

			public bool Contains(State state)
			{
				for (int i = 0; i < channels.Length; i++)
					if (state.channels[i] && !channels[i])
						return false;
				return true;
			}

			public int Contains(IList<State> finalStates)
			{
				for (int i = 0; i < finalStates.Count; i++)
					if (this.Contains(finalStates[i]))
						return i;
				return -1;
			}
		}

		private class StateSet
		{
			private IList<State> states;

			public IList<State> States { get { return states; } }

			public StateSet()
			{ 
				this.states = new List<State>();
			}

			public State Find(State state)
			{
				int index = states.IndexOf(state);
				if (index >= 0)
					return states[index];
				else {
					Debug.Assert(!state.Initialized);
					state.Index = states.Count;
					states.Add(state);
					return state;
				}
			}
		}

    private static State GetStateOfRow(JoinInputDefinition.Row row, int nChannels)
    {
      State state = new State(nChannels);
      foreach (JoinInputDefinition.Atom atom in row.Atoms)
				state = state.Full(atom.Index);
      return state;
    }

    public JoinAutomaton(IList<JoinInputDefinition.Row> rows)
    {
      this.state = 0;

			// compute the number of different channels involved
			nChannels = 0;
			foreach (JoinInputDefinition.Row row in rows)
				foreach (JoinInputDefinition.Atom atom in row.Atoms)
					nChannels = Math.Max(nChannels, atom.Index + 1);
      Debug.Assert(nChannels > 0);

			// compute the set of final states
      IList<State> finalStates = new List<State>();
      foreach (JoinInputDefinition.Row row in rows) {
				State finalState = GetStateOfRow(row, nChannels);
				Debug.Assert(!finalStates.Contains(finalState));
				finalStates.Add(finalState);
      }

			// compute the set of states
      StateSet stateSet = new StateSet();
      Queue<State> stateQueue = new Queue<State>();
      stateQueue.Enqueue(new State(nChannels));
      while (stateQueue.Count > 0) {
				State state = stateSet.Find(stateQueue.Dequeue());
				if (!state.Initialized) {
				  state.Initialized = true;
				  state.Final = state.Contains(finalStates);
				  state.Transitions = new int[2 * nChannels];
				  for (int i = 0; i < nChannels; i++) {
				    if (state.Final > 0)
				      state.Transitions[nChannels + i] = 0;
	    			else {
				      State newStatePositive = stateSet.Find(state.Full(i));
				      state.Transitions[nChannels + i] = newStatePositive.Index;
	    			  stateQueue.Enqueue(newStatePositive);
				    }

	    			State newStateNegative = stateSet.Find(state.Empty(i));
				    state.Transitions[i] = newStateNegative.Index;
		  		  stateQueue.Enqueue(newStateNegative);
				  }
				}
      }

			// compute the transition matrix and the continuations
			int nStates = stateSet.States.Count;
			transitions = new int[nStates, 2 * nChannels];
			continuations = new int[nStates];
			foreach (State state in stateSet.States) {
				Debug.Assert(state.Index >= 0 && state.Index < nStates);
				for (int i = 0; i < nChannels; i++) {
					// TODO: EFFICIENT COPY
					transitions[state.Index, i] = state.Transitions[i];
					transitions[state.Index, nChannels + i] = state.Transitions[nChannels + i];
				}
				continuations[state.Index] = state.Final;
			}

#if false
			Console.WriteLine("automaton has {0} states", nStates);
			for (int i = 0; i < nStates; i++) {
				for (int j = 0; j < nChannels; j++) {
					Console.WriteLine("{0} -{1} => {2}", i, j, transitions[i, j]);
					Console.WriteLine("{0} +{1} => {2}", i, j, transitions[i, nChannels + j]);
				}
				Console.WriteLine("continuation: {0}", continuations[i]);
			}
#endif
    }

		public int ActiveRow { get { return continuations[state]; } }

		public void Full(int channelIndex)
		{
			Debug.Assert(channelIndex >= 0 && channelIndex < this.nChannels);
      state = transitions[state, this.nChannels + channelIndex];
		}

    public void Empty(int channelIndex)
    {
			Debug.Assert(channelIndex >= 0 && channelIndex < this.nChannels);
      state = transitions[state, channelIndex];
		}
	}

  public class JoinInputDefinition
  {
		private readonly bool service;
		private readonly IList<Row> rows;
		private readonly IJoinAutomaton automaton;
		private readonly IList<JoinInputRequest> requests;
    private readonly Env env;
    private readonly VmThread thread;

		public class Atom
		{
			private readonly Channel channel;
			private readonly int index;
			private readonly IPattern pattern;

			public Channel Channel { get { return channel; } }
			public int Index { get { return index; } }
			public IPattern Pattern { get { return pattern; } }

			public Atom(Channel channel, int index, IPattern pattern)
			{
				this.index = index;
				this.channel = channel;
				this.pattern = pattern;
			}
		}

		public class Row
		{
			private readonly IList<Atom> atoms;
			private readonly XmlElement continuation;

			public IList<Atom> Atoms { get { return atoms; } }
			public XmlElement Continuation { get { return continuation; } }

			public Row(IList<Atom> atoms, XmlElement continuation)
			{
				this.atoms = atoms;
				this.continuation = continuation;
			}

			public void Match(Env env)
			{
				foreach (Atom atom in atoms) {
					IValue thisValue = atom.Channel.DequeueValue();
					bool matchRes = MatcherFactory.Make(atom.Pattern).Match(thisValue, env) >= 0;
					Debug.Assert(matchRes);
				}
			}

			public void Match(Env env, int channelIndex, IValue v)
			{
				foreach (Atom atom in atoms) {
					IValue thisValue = (channelIndex == atom.Index) ? v : atom.Channel.DequeueValue();
					bool matchRes = MatcherFactory.Make(atom.Pattern).Match(thisValue, env) >= 0;
					Debug.Assert(matchRes);
				}
			}
		}

    public JoinInputDefinition(bool service,
															 IList<Channel> channels,
															 IList<Row> rows,
															 IJoinAutomaton automaton,
			       									 Env env,
			       									 VmThread thread)
    {
			this.service = service;
			this.rows = rows;
			this.automaton = automaton;
      this.env = env;
      this.thread = thread;

			// WARNING: the atom index is computed by the compiler and stored in the
			// object code. However input requests recompute this index according to
			// the first occurrence of a channel in the channels of a join
			// definition. As long as this is the same strategy adopted by the
			// compiler for computing atom indexes it all fine but this should be
			// eventually patched
      Debug.Assert(channels.Count > 0);
      requests = new List<JoinInputRequest>();
      for (int i = 0; i < channels.Count; i++) {
				JoinInputRequest request = new JoinInputRequest(this, channels[i], i);
				requests.Add(request);
				channels[i].EnqueueRequest(request);
			}
    }

    public bool Send(JoinInputRequest request, IValue v)
    {
      if (request.Channel.Empty()) {
				automaton.Full(request.Index);
				int row = automaton.ActiveRow;
				if (row >= 0) {
					// the channel was empty when this method was called, we declared
					// it full but never really pushed the value in its queue. Now
					// there is a match, which means that the value v has activated a
					// pattern and it is consumed immediately. Thus, we declare that
					// the channel is empty now
					automaton.Empty(request.Index);
					Debug.Assert(request.Channel.Empty());

					if (service) {
						Env newEnv = env.Clone();
	  				rows[row].Match(newEnv, request.Index, v);
						thread.Fork(rows[row].Continuation, newEnv);
					} else {
						rows[row].Match(env, request.Index, v);
						foreach (JoinInputRequest req in requests)
							req.Channel.RemoveRequest(req);
						thread.Unblock(rows[row].Continuation);
					}

  				return true;
				}
      }

      return false;
    }

		public void Empty(JoinInputRequest request)
		{ automaton.Empty(request.Index); }
  }
}
