// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.IO;
using System.Xml;
using System.Diagnostics;
using System.Collections.Generic;
using PiDuce.Common;
using PiDuce.Machine;
using PiDuce.Web;

namespace PiDuce.Machine
{
  public interface IValue
  {
    void Accept(IValueVisitor visitor);
		ISchema SchemaOf();
  }

  public interface IValueVisitor
  {
    void VisitVoidValue(VoidValue v);
    void VisitIntValue(IntValue v);
    void VisitFloatValue(FloatValue v);
    void VisitBoolValue(BoolValue v);
    void VisitStringValue(StringValue v);
    void VisitChannelValue(ChannelValue v);
    void VisitServiceValue(ServiceValue v);
    void VisitLabelValue(LabelValue v);
    void VisitSequenceValue(SequenceValue v);
  }

	public abstract class Value : IValue
	{
    public abstract void Accept(IValueVisitor visitor);
		public abstract ISchema SchemaOf();

		public static IValue MakeValueFromLiteral(IBasicTypeLiteral literal)
		{
			switch (literal.TypeOf().Kind) {
				case BasicTypeKind.BOOL:
					return new BoolValue(((BoolLiteral) literal).Value);
				case BasicTypeKind.INT:
					return new IntValue(((IntLiteral) literal).Value);
				case BasicTypeKind.FLOAT:
					return new FloatValue(((FloatLiteral) literal).Value);
				case BasicTypeKind.STRING:
					return new StringValue(((StringLiteral) literal).Value);
				default:
					Debug.Assert(false);
					return null;
			}
		}
	}

	public class VoidValue : Value
	{
    public override void Accept(IValueVisitor visitor)
    { visitor.VisitVoidValue(this); }

		public override string ToString()
		{ return "void"; }


		public override ISchema SchemaOf()
		{ return new VoidSchema(); }
	}

	public abstract class BasicValue<T> : Value
	{
		private readonly T value;

		public T Value { get { return this.value; } }

		public BasicValue(T value)
		{ this.value = value; }

		public override string ToString()
		{ return value.ToString(); }

		public override bool Equals(Object obj)
		{
			if (obj is BasicValue<T>)
				return Value.Equals(((BasicValue<T>) obj).Value);
			else
				return false;
		}

		public override int GetHashCode()
		{ return value.GetHashCode(); }
	}

  public class BoolValue : BasicValue<bool>
  { 
    public BoolValue(bool value) : base(value)
		{ }

		public override void Accept(IValueVisitor visitor)
    { visitor.VisitBoolValue(this); }

    public static BoolValue operator !(BoolValue op)
    { return new BoolValue(!op.Value); }


		public override ISchema SchemaOf()
		{ return new BasicSchemaLiteralSchema(new BoolLiteral(this.Value)); }

		public override string ToString()
		{ return Value.ToString(); }
  }

	public class IntValue : BasicValue<int>
	{
		public IntValue(int value) : base(value)
		{ }

		public override void Accept(IValueVisitor visitor)
    { visitor.VisitIntValue(this); }

		public override string ToString()
		{ return Value.ToString(); }

		public override ISchema SchemaOf()
		{ return new BasicSchemaLiteralSchema(new IntLiteral(Value)); }

	}

	public class FloatValue : BasicValue<float>
	{
		public FloatValue(float value) : base(value)
		{ }

    public override void Accept(IValueVisitor visitor)
    { visitor.VisitFloatValue(this); }

		public override string ToString()
		{ return Value.ToString(); }

		public override ISchema SchemaOf()
		{ return new BasicSchemaLiteralSchema(new FloatLiteral(Value)); }
	}

	public class StringValue : BasicValue<string>
	{
		public StringValue(string value) : base(value)
		{ }

    public override void Accept(IValueVisitor visitor)
    { visitor.VisitStringValue(this); }

		public override string ToString()
		{ return Value; }


		public override ISchema SchemaOf()
		{ return new BasicSchemaLiteralSchema(new StringLiteral(Value)); }
	}

  public abstract class ChannelValue : Value
  {
    private readonly string location;
    private readonly string operationName;
    private readonly string wsdlLocation;
    private readonly ISchema schema;

    protected ChannelValue() { }

    public ChannelValue(string wsdlLocation, string operationName, string location, ISchema schema)
    {
      Debug.Assert(schema != null);
      Debug.Assert(operationName != null && operationName != String.Empty);
      Debug.Assert(schema is ChannelSchema || schema is FunctionSchema);
      this.location = location;
      this.operationName = operationName;
      this.schema = schema;
      this.wsdlLocation = wsdlLocation;
    }

    public string Location { get { return location; } }
    public string OperationName { get { return operationName; } }
    public string WsdlLocation { get { return wsdlLocation; } }
    public ISchema Schema { get { return schema; } }

		public override ISchema SchemaOf()
		{ return Schema; }
		// LUCA: is this the schema of the channel or the schema of the
		// values carried by the channel? I'd prefer the latter, but it seems
		// like it's the former


    public override void Accept(IValueVisitor visitor)
    { visitor.VisitChannelValue(this); }

    public override string ToString()
    { return "chan(" + wsdlLocation + "," + operationName + ")"; }

    public abstract bool IsLocalChannel();
    public abstract bool IsRemoteChannel();
    public abstract LocalChannel AsLocalChannel();
    public abstract RemoteChannel AsRemoteChannel();
  }

  public class LocalChannel : ChannelValue
  {
    public LocalChannel(string wsdlLocation, string operationName, string location, ISchema schema) 
      : base(wsdlLocation, operationName, location, schema) { }
    public override bool IsLocalChannel() { return true; }
    public override bool IsRemoteChannel() { return false; }
    public override LocalChannel AsLocalChannel() { return this; }
    public override RemoteChannel AsRemoteChannel() { throw new ApplicationException("Cannot cast from LocalChannel to RemoteChannel"); }
  }

	public class RemoteChannel : ChannelValue
	{
    private Operation operation;

    public string NamespaceUri { get { return operation.NamespaceUri; } }
    public SoapOverHttpBinding Binding { get { return (SoapOverHttpBinding)operation.Binding; } }
    public Operation Operation { get { return operation; } }

    public RemoteChannel(Operation operation) : base(operation.WsdlLocation, operation.Name, operation.ChanLocation, operation.Schema)
		{
      this.operation = operation; 
		}
    public override bool IsLocalChannel() { return false; }
    public override bool IsRemoteChannel() { return true; }
    public override LocalChannel AsLocalChannel() { throw new ApplicationException("Cannot cast from RemoteChannel to LocalChannel"); }
    public override RemoteChannel AsRemoteChannel() { return this; }
	}

  public abstract class ServiceValue : Value
  {
    protected readonly IDictionary<string, ChannelValue> operations;
    public IDictionary<string, ChannelValue> Operations { get { return this.operations; } }

    protected readonly string wsdlLocation;
    public string WsdlLocation { get { return wsdlLocation; } }

    public ServiceValue(string wsdlLocation)
    { 
      this.wsdlLocation = wsdlLocation;
      this.operations = new Dictionary<string, ChannelValue>();
    }

    public override void Accept(IValueVisitor visitor)
    { visitor.VisitServiceValue(this); }

    public override string ToString()
    { return "service(" + wsdlLocation + ")"; }

		public override ISchema SchemaOf()
		{
			IDictionary<string, ISchema> t = new Dictionary<string, ISchema>();
			foreach (KeyValuePair<string, ChannelValue> p in operations)
				t[p.Key] = p.Value.SchemaOf();
			return new ServiceSchema(t);
		}
  }

  public class LocalService : ServiceValue
  {
    public LocalService(string wsdlLocation, IDictionary<string, ISchema> operations) 
      : base(wsdlLocation)
    {
      foreach (KeyValuePair<string, ISchema> op in operations)
      {
        this.operations[op.Key] = new LocalChannel(wsdlLocation, op.Key, wsdlLocation.ToString().Substring(0, wsdlLocation.ToString().IndexOf("?")) + "/" + op.Key, op.Value);
      }
    }
  }

  public class RemoteService : ServiceValue
  {
    public RemoteService(string wsdlLocation) : 
      this (wsdlLocation, Web.Webclient.Download(wsdlLocation))
    {  }

    public RemoteService(string wsdlLocation, string wsdl) : base (wsdlLocation)
    {
      WsdlReader decoder = new WsdlReader();
      decoder.Compile(wsdl, wsdlLocation);
      IDictionary<string, Operation> operations = decoder.Operations;
      foreach (Operation operation in operations.Values)
      { this.operations[operation.Name] = new RemoteChannel(operation); }
    }

    public override string ToString()
    { return "TODO"; }
  }

	public class LabelValue : Value
	{
		private readonly IValue content;
		private readonly string label;
		private readonly IDictionary<string, IValue> attributes;

		public IValue Content { get { return content; } }
		public string Label { get { return label; } }
		public IDictionary<string, IValue> Attributes { get { return attributes; } } 

		public LabelValue(string label, IValue content, IDictionary<string, IValue> attributes) 
		{
			this.label = label;
			this.content = content;
			this.attributes = attributes;
		}		

		public LabelValue(string label, IValue content) :
			this(label, content, new Dictionary<string, IValue>())
		{		}
						
    public override void Accept(IValueVisitor visitor)
    {
      visitor.VisitLabelValue(this);
    }
		
		public override string ToString()
		{ return label + "[" + content.ToString() + "]"; }

		public override ISchema SchemaOf()
		{ 
			IDictionary<string, ISchema> temp = new Dictionary<string, ISchema>();
			foreach(KeyValuePair<string, IValue> pair in attributes)
				temp.Add(pair.Key, pair.Value.SchemaOf());
			return new LabelledSchema(LabelSet.Singleton(label), content.SchemaOf(), new RecordSchema(temp)); 
		}
	}

	public class SequenceValue : Value
	{
		private readonly IValue top;
		private readonly IValue tail;

		public IValue Top { get { return top; } }
		public IValue Tail { get { return tail; } }

		private SequenceValue(IValue top, IValue tail)
		{
			this.top = top;
			this.tail = tail;
		}

    public int Length { 
      get 
      {
        Debug.Assert(!(top is SequenceValue));
        if (Tail is SequenceValue)
          return 1 + ((SequenceValue)Tail).Length;
        else return 2;
      } 
    }

		public static IValue Make(IValue head, IValue tail)
		{
			if (tail is VoidValue)
				return head;
			else if (head is VoidValue)
				return tail;
			else if (head is SequenceValue)
				return Make(((SequenceValue) head).Top, Make(((SequenceValue) head).Tail, tail));
			else
				return new SequenceValue(head, tail);
		}

    public override void Accept(IValueVisitor visitor)
    {
      visitor.VisitSequenceValue(this);
    }

		public override string ToString()
		{ return top.ToString() + "," + tail.ToString(); }


		public override ISchema SchemaOf()
		{ return new SequenceSchema(top.SchemaOf(), tail.SchemaOf()); }
	}
}

